@@ -35,13 +35,16 @@ import os
 os.environ['GRL_PLUGIN_RANKS'] = ("grl-local-metadata:5,"
                                   "grl-filesystem:4,"
                                   "grl-tracker-source:3,"
-                                  "grl-lastfm-cover:2")
+                                  "grl-lastfm-cover:2,"
+                                  "grl-theaudiodb-cover:1")
 from gi.repository import Grl
 logger = logging.getLogger(__name__)
 
 
 class Grilo(GObject.GObject):
 
+    __gtype_name__ = 'Grilo'
+
     __gsignals__ = {
         'ready': (GObject.SignalFlags.RUN_FIRST, None, ()),
         'changes-pending': (GObject.SignalFlags.RUN_FIRST, None, ()),
@@ -74,12 +77,18 @@ class Grilo(GObject.GObject):
     CHANGED_MEDIA_MAX_ITEMS = 500
     CHANGED_MEDIA_SIGNAL_TIMEOUT = 2000
 
+    _theaudiodb_api_key = "195003"
+
+    sources = GObject.Property()
+    cover_sources = GObject.Property(type=bool, default=False)
+
     def __repr__(self):
         return '<Grilo>'
 
     @log
     def __init__(self):
-        GObject.GObject.__init__(self)
+        super().__init__()
+
         self.playlist_path = GLib.build_filenamev([GLib.get_user_data_dir(),
                                                   "gnome-music", "playlists"])
         if not (GLib.file_test(self.playlist_path, GLib.FileTest.IS_DIR)):
@@ -94,7 +103,8 @@ class Grilo(GObject.GObject):
         self.full_options.set_resolution_flags(Grl.ResolutionFlags.FULL |
                                                Grl.ResolutionFlags.IDLE_RELAY)
 
-        self.sources = {}
+        self.props.sources = {}
+
         self.blacklist = [
             'grl-filesystem',
             'grl-bookmarks',
@@ -108,12 +118,21 @@ class Grilo(GObject.GObject):
         self.changes_pending = {'Albums': False, 'Artists': False, 'Songs': False}
         self.pending_changed_medias = []
 
+        self._thumbnail_sources = []
+        self._thumbnail_sources_timeout = None
+
         self.registry = Grl.Registry.get_default()
 
         self.sparqltracker = TrackerWrapper().tracker
 
+        self._find_sources()
+
     @log
     def _find_sources(self):
+        config = Grl.Config.new('grl-lua-factory', 'grl-theaudiodb-cover')
+        config.set_api_key(self._theaudiodb_api_key)
+        self.registry.add_config(config)
+
         self.registry.connect('source_added', self._on_source_added)
         self.registry.connect('source_removed', self._on_source_removed)
 
@@ -124,6 +143,7 @@ class Grilo(GObject.GObject):
         if self.tracker is not None:
             logger.debug("tracker found")
 
+    @log
     def _rate_limited_content_changed(self, mediaSource, changedMedias, changeType, locationUnknown):
         [self.pending_changed_medias.append(media) for media in changedMedias]
         if self.content_changed_timeout is None:
@@ -136,22 +156,17 @@ class Grilo(GObject.GObject):
             with self.tracker.handler_block(self.notification_handler):
                 for media in changedMedias:
                     media_id = media.get_id()
-                    if changeType == Grl.SourceChangeType.ADDED:
-                        # Check that this media is an audio file
-                        mime_type = self.tracker.query_sync(
-                            Query.is_audio(media_id),
-                            [Grl.METADATA_KEY_MIME],
-                            self.options)[0].get_mime()
-                        if mime_type and mime_type.startswith("audio"):
-                            self.changed_media_ids.append(media_id)
+                    if (changeType == Grl.SourceChangeType.ADDED
+                            and media.is_audio()):
+                        self.changed_media_ids.append(media_id)
                     if changeType == Grl.SourceChangeType.REMOVED:
                         # There is no way to check that removed item is a media
                         # so always do the refresh
                         # todo: remove one single url
                         try:
                             self.changed_media_ids.append(media.get_id())
-                        except Exception as e:
-                            logger.warn("Skipping %s", media)
+                        except Exception:
+                            logger.warning("Skipping {}".format(media))
 
                 if self.changed_media_ids == []:
                     self.pending_changed_medias = []
@@ -170,8 +185,9 @@ class Grilo(GObject.GObject):
                         GLib.Source.remove(self.pending_event_id)
                         self.pending_event_id = 0
                     self.pending_event_id = GLib.timeout_add(self.CHANGED_MEDIA_SIGNAL_TIMEOUT, self.emit_change_signal)
-        except Exception as e:
-            logger.warn("Exception in _on_content_changed: %s", e)
+        except Exception as error:
+            logger.warning(
+                "Exception in _on_content_changed: {}".format(error))
         finally:
             self.pending_changed_medias = []
             if self.content_changed_timeout is not None:
@@ -189,6 +205,14 @@ class Grilo(GObject.GObject):
         self.emit('changes-pending')
         return False
 
+    @log
+    def _trigger_art_update(self):
+        self._thumbnail_sources_timeout = None
+        if len(self._thumbnail_sources) > 0:
+            self.props.cover_sources = True
+
+        return GLib.SOURCE_REMOVE
+
     @log
     def _on_source_added(self, pluginRegistry, mediaSource):
         if ("net:plaintext" in mediaSource.get_tags()
@@ -196,10 +220,18 @@ class Grilo(GObject.GObject):
             try:
                 pluginRegistry.unregister_source(mediaSource)
             except GLib.GError:
-                logger.error("Failed to unregister %s.",
-                             mediaSource.get_id())
+                logger.error(
+                    "Failed to unregister {}".format(mediaSource.get_id()))
             return
 
+        if Grl.METADATA_KEY_THUMBNAIL in mediaSource.supported_keys():
+            self._thumbnail_sources.append(mediaSource)
+            if not self._thumbnail_sources_timeout:
+                # Aggregate sources being added, for example when the
+                # network comes online.
+                self._thumbnail_sources_timeout = GLib.timeout_add_seconds(
+                    5, self._trigger_art_update)
+
         id = mediaSource.get_id()
         logger.debug("new grilo source %s was added", id)
         try:
@@ -208,7 +240,7 @@ class Grilo(GObject.GObject):
             if id == 'grl-tracker-source':
                 if ops & Grl.SupportedOps.SEARCH:
                     logger.debug("found searchable tracker source")
-                    self.sources[id] = mediaSource
+                    self.props.sources[id] = mediaSource
                     self.tracker = mediaSource
                     self.search_source = mediaSource
 
@@ -221,17 +253,17 @@ class Grilo(GObject.GObject):
 
             elif (id.startswith('grl-upnp')):
                 logger.debug("found upnp source %s", id)
-                self.sources[id] = mediaSource
+                self.props.sources[id] = mediaSource
                 self.emit('new-source-added', mediaSource)
 
             elif (ops & Grl.SupportedOps.SEARCH
                   and mediaSource.get_supported_media() & Grl.MediaType.AUDIO):
                 logger.debug("source %s is searchable", id)
-                self.sources[id] = mediaSource
+                self.props.sources[id] = mediaSource
                 self.emit('new-source-added', mediaSource)
 
         except Exception as e:
-            logger.debug("Source %s: exception %s", id, e)
+            logger.debug("Source {}: exception {}".format(id, e))
 
     @log
     def _on_source_removed(self, pluginRegistry, mediaSource):
@@ -239,35 +271,44 @@ class Grilo(GObject.GObject):
 
     @log
     def populate_artists(self, offset, callback, count=-1):
-        if self.tracker:
-            GLib.idle_add(self.populate_items, Query.all_artists(), offset,
-                          callback, count)
+        self.populate_items(Query.all_artists(), offset, callback, count)
 
     @log
     def populate_albums(self, offset, callback, count=-1):
-        if self.tracker:
-            GLib.idle_add(self.populate_items, Query.all_albums(), offset,
-                                                callback, count)
+        self.populate_items(Query.all_albums(), offset, callback, count)
 
     @log
     def populate_songs(self, offset, callback, count=-1):
-        if self.tracker:
-            GLib.idle_add(self.populate_items, Query.all_songs(), offset,
-                                                callback, count)
+        self.populate_items(Query.all_songs(), offset, callback, count)
 
     @log
     def populate_playlists(self, offset, callback, count=-1):
-        if self.tracker:
-            GLib.idle_add(self.populate_items, Query.all_playlists(), offset,
-                                                callback, count)
+        """Asynchronously get playlists (user and smart ones)
+
+        :param int offset: start index
+        :param function callback: callback function
+        :param int count: limit number of results
+        """
+        self.populate_items(Query.all_playlists(), offset, callback, count)
+
+    @log
+    def populate_user_playlists(self, offset, callback, count=-1):
+        """Asynchronously get user playlists
+
+        :param int offset: start index
+        :param function callback: callback function
+        :param int count: limit number of results
+        """
+        self.populate_items(
+            Query.all_user_playlists(), offset, callback, count)
 
     @log
     def populate_album_songs(self, album, callback, count=-1):
         if album.get_source() == 'grl-tracker-source':
-            GLib.idle_add(self.populate_items,
-                          Query.album_songs(album.get_id()), 0, callback, count)
+            self.populate_items(
+                Query.album_songs(album.get_id()), 0, callback, count)
         else:
-            source = self.sources[album.get_source()]
+            source = self.props.sources[album.get_source()]
             length = len(album.songs)
             for i, track in enumerate(album.songs):
                 callback(source, None, track, length - (i + 1), None)
@@ -275,10 +316,8 @@ class Grilo(GObject.GObject):
 
     @log
     def populate_playlist_songs(self, playlist, callback, count=-1):
-        if self.tracker:
-            GLib.idle_add(self.populate_items,
-                          Query.playlist_songs(playlist.get_id()), 0, callback,
-                          count)
+        self.populate_items(
+            Query.playlist_songs(playlist.get_id()), 0, callback, count)
 
     @log
     def populate_custom_query(self, query, callback, count=-1, data=None):
@@ -292,6 +331,9 @@ class Grilo(GObject.GObject):
             options.set_count(count)
 
         def _callback(source, param, item, remaining, data, error):
+            if error:
+                logger.warning(
+                    "Error {}: {}".format(error.domain, error.message))
             callback(source, param, item, remaining, data)
         self.tracker.query(query, self.METADATA_KEYS, options, _callback, data)
 
@@ -312,9 +354,11 @@ class Grilo(GObject.GObject):
 
         # FIXME: We assume this is the tracker plugin.
         # FIXME: Doing this async crashes
-        self.tracker.store_metadata_sync(song_item,
-                                         [Grl.METADATA_KEY_FAVOURITE],
-                                         Grl.WriteFlags.NORMAL)
+        try:
+            self.tracker.store_metadata_sync(
+                song_item, [Grl.METADATA_KEY_FAVOURITE], Grl.WriteFlags.NORMAL)
+        except GLib.Error as error:
+            logger.warning("Error {}: {}".format(error.domain, error.message))
 
     @log
     def set_favorite(self, song_item, favorite):
@@ -333,11 +377,17 @@ class Grilo(GObject.GObject):
 
         @log
         def _search_callback(source, param, item, remaining, data, error):
+            if error:
+                logger.warning(
+                    "Error {}: {}".format(error.domain, error.message))
             callback(source, param, item, remaining, data)
             self._search_callback_counter += 1
 
         @log
         def _multiple_search_callback(source, param, item, remaining, data, error):
+            if error:
+                logger.warning(
+                    "Error {}: {}".format(error.domain, error.message))
             callback(source, param, item, remaining, data)
 
         if self.search_source:
@@ -346,13 +396,17 @@ class Grilo(GObject.GObject):
             self.search_source.search(q, self.METADATA_KEYS, options,
                                       _search_callback, data)
         else:
-            Grl.multiple_search([self.sources[key] for key in self.sources
-                                 if key != 'grl-tracker-source'],
-                                q, self.METADATA_KEYS, options,
-                                _multiple_search_callback, data)
+            Grl.multiple_search(
+                [self.props.sources[key] for key in self.props.sources
+                 if key != 'grl-tracker-source'], q,
+                 self.METADATA_KEYS, options, _multiple_search_callback, data)
 
     @log
     def get_album_art_for_item(self, item, callback):
+        # Tracker not (yet) loaded.
+        if self.search_source is None:
+            return
+
         item_id = item.get_id()
 
         if item.is_audio():
@@ -392,8 +446,11 @@ class Grilo(GObject.GObject):
 
         # FIXME: We assume this is the tracker plugin.
         # FIXME: Doing this async crashes
-        self.tracker.store_metadata_sync(media, [Grl.METADATA_KEY_PLAY_COUNT],
-                                         Grl.WriteFlags.NORMAL)
+        try:
+            self.tracker.store_metadata_sync(
+                media, [Grl.METADATA_KEY_PLAY_COUNT], Grl.WriteFlags.NORMAL)
+        except GLib.Error as error:
+            logger.warning("Error {}: {}".format(error.domain, error.message))
 
     @log
     def set_last_played(self, media):
@@ -405,8 +462,11 @@ class Grilo(GObject.GObject):
         media.set_last_played(GLib.DateTime.new_now_utc())
         # FIXME: We assume this is the tracker plugin.
         # FIXME: Doing this async crashes
-        self.tracker.store_metadata_sync(media, [Grl.METADATA_KEY_LAST_PLAYED],
-                                         Grl.WriteFlags.NORMAL)
+        try:
+            self.tracker.store_metadata_sync(
+                media, [Grl.METADATA_KEY_LAST_PLAYED], Grl.WriteFlags.NORMAL)
+        except GLib.Error as error:
+            logger.warning("Error {}: {}".format(error.domain, error.message))
 
     @log
     def songs_available(self, callback):
@@ -420,7 +480,7 @@ class Grilo(GObject.GObject):
             try:
                 has_next = conn.next_finish(res)
             except GLib.Error as err:
-                logger.warn("Error: %s, %s", err.__class__, err)
+                logger.warning("Error: {}, {}".format(err.__class__, err))
                 callback(False)
                 return
 
@@ -437,7 +497,7 @@ class Grilo(GObject.GObject):
             try:
                 cursor = conn.query_finish(res)
             except GLib.Error as err:
-                logger.warn("Error: %s, %s", err.__class__, err)
+                logger.warning("Error: {}, {}".format(err.__class__, err))
                 callback(False)
                 return
 
@@ -448,45 +508,5 @@ class Grilo(GObject.GObject):
         self.sparqltracker.query_async(Query.all_songs_count(), None,
                                        songs_query_cb, None)
 
-    @log
-    def playlists_available(self, callback):
-        """Checks if there are any non-static playlists available
-
-        Calls a callback function with True or False depending on the
-        availability of playlists.
-        :param callback: Function to call on result
-        """
-        def cursor_next_cb(conn, res, data):
-            try:
-                has_next = conn.next_finish(res)
-            except GLib.Error as err:
-                logger.warn("Error: %s, %s", err.__class__, err)
-                callback(False)
-                return
-
-            if has_next:
-                count = conn.get_integer(0)
-
-                if count > 0:
-                    callback(True)
-                    return
-
-            callback(False)
-
-        def playlists_query_cb(conn, res, data):
-            try:
-                cursor = conn.query_finish(res)
-            except GLib.Error as err:
-                logger.warn("Error: %s, %s", err.__class__, err)
-                callback(False)
-                return
-
-            cursor.next_async(None, cursor_next_cb, None)
-
-        # TODO: currently just checks tracker, should work with any
-        # queryable supported Grilo source.
-        self.sparqltracker.query_async(
-                Query.all_non_static_playlists_count(),
-                None, playlists_query_cb, None)
 
 grilo = Grilo()
